=== modified file 'src/CMakeLists.txt'
--- src/CMakeLists.txt	2012-01-20 15:42:52 +0000
+++ src/CMakeLists.txt	2012-01-21 13:34:31 +0000
@@ -99,6 +99,7 @@
     ${CMAKE_CURRENT_SOURCE_DIR}/modifierhandler.cpp
     ${CMAKE_CURRENT_SOURCE_DIR}/propertywriter.cpp
     ${CMAKE_CURRENT_SOURCE_DIR}/eventsource.cpp
+    ${CMAKE_CURRENT_SOURCE_DIR}/signalsource.cpp
     ${CMAKE_CURRENT_SOURCE_DIR}/stackdebugger.cpp
     
     ${_bcop_sources}

=== modified file 'src/eventsource.cpp'
--- src/eventsource.cpp	2012-01-20 17:50:31 +0000
+++ src/eventsource.cpp	2012-01-21 13:34:31 +0000
@@ -60,13 +60,7 @@
 bool
 CompEventSource::callback ()
 {
-    if (restartSignal || shutDown)
-    {
-	screen->priv->mainloop->quit ();
-	return false;
-    }
-    else
-	screen->priv->processEvents ();
+    screen->priv->processEvents ();
     return true;
 }
 

=== modified file 'src/main.cpp'
--- src/main.cpp	2012-01-19 18:12:31 +0000
+++ src/main.cpp	2012-01-21 13:34:31 +0000
@@ -57,7 +57,7 @@
 }
 
 static void
-signalHandler (int sig)
+chldSignalHandler (int sig)
 {
     int status;
 
@@ -65,14 +65,6 @@
     case SIGCHLD:
 	waitpid (-1, &status, WNOHANG | WUNTRACED);
 	break;
-    case SIGHUP:
-	restartSignal = true;
-	break;
-    case SIGINT:
-    case SIGTERM:
-	shutDown = true;
-    default:
-	break;
     }
 }
 
@@ -248,10 +240,7 @@
     programArgc = argc;
     programArgv = argv;
 
-    signal (SIGHUP, signalHandler);
-    signal (SIGCHLD, signalHandler);
-    signal (SIGINT, signalHandler);
-    signal (SIGTERM, signalHandler);
+    signal (SIGCHLD, chldSignalHandler);
 
     if (!manager.parseArguments (argc, argv))
 	return 0;

=== modified file 'src/privatescreen.h'
--- src/privatescreen.h	2012-01-20 17:50:31 +0000
+++ src/privatescreen.h	2012-01-21 13:34:31 +0000
@@ -41,6 +41,7 @@
 #include "privatetimeoutsource.h"
 #include "privateiosource.h"
 #include "privateeventsource.h"
+#include "privatesignalsource.h"
 
 #include "core_options.h"
 
@@ -382,6 +383,8 @@
 	
 	void setDefaultWindowAttributes (XWindowAttributes *);
 
+	void handleSignal (int signum);
+
     public:
 
 	PrivateScreen *priv;
@@ -393,6 +396,9 @@
 	 */
 	CompEventSource * source;
 	CompTimeoutSource * timeout;
+	CompSignalSource * sighupSource;
+	CompSignalSource * sigtermSource;
+	CompSignalSource * sigintSource;
 	Glib::RefPtr <Glib::MainContext> ctx;
 
 	CompFileWatchList   fileWatch;

=== added file 'src/privatesignalsource.h'
--- src/privatesignalsource.h	1970-01-01 00:00:00 +0000
+++ src/privatesignalsource.h	2012-01-21 13:34:31 +0000
@@ -0,0 +1,55 @@
+/*
+ * Copyright © 2012 Canonical, Ltd.
+ *
+ * Permission to use, copy, modify, distribute, and sell this software
+ * and its documentation for any purpose is hereby granted without
+ * fee, provided that the above copyright notice appear in all copies
+ * and that both that copyright notice and this permission notice
+ * appear in supporting documentation, and that the name of
+ * Canonical, Ltd. not be used in advertising or publicity pertaining to
+ * distribution of the software without specific, written prior permission.
+ * Canonical, Ltd. makes no representations about the suitability of this
+ * software for any purpose. It is provided "as is" without express or
+ * implied warranty.
+ *
+ * CANONICAL LTD. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
+ * NO EVENT SHALL Canonical, Ltd. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
+ * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
+ * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
+ * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ *
+ * Author: Sam Spilsbury <sam.spilsbury@canonical.com>
+ */
+
+#include <boost/function.hpp>
+
+#include <glib.h>
+
+class CompSignalSource
+{
+    public:
+
+	~CompSignalSource ();
+
+	typedef boost::function <void (int)> callbackFunc;
+
+	static
+	CompSignalSource *
+	create (int signum,
+		const callbackFunc &);
+
+    protected:
+
+	explicit CompSignalSource (int signum, const callbackFunc &);
+
+    private:
+
+	static gboolean callback (gpointer user_data);
+	static void     destroyed (gpointer user_data);
+
+	callbackFunc mFunc;
+	int          mSignal;
+	gint         mSource;
+};

=== modified file 'src/screen.cpp'
--- src/screen.cpp	2012-01-20 18:12:12 +0000
+++ src/screen.cpp	2012-01-21 13:34:31 +0000
@@ -112,10 +112,28 @@
 }
 
 void
+PrivateScreen::handleSignal (int signum)
+{
+    switch (signum)
+    {
+	case SIGINT:
+	case SIGTERM:
+	    shutDown = true;
+	    break;
+	case SIGHUP:
+	    restartSignal = true;
+	    break;
+	default:
+	    break;
+    }
+
+    if (shutDown || restartSignal)
+	mainloop->quit ();
+}
+
+void
 CompScreen::eventLoop ()
 {
-    priv->ctx = Glib::MainContext::get_default ();
-    priv->mainloop = Glib::MainLoop::create (priv->ctx, false);
     priv->source = CompEventSource::create ();
     priv->timeout = CompTimeoutSource::create (priv->ctx);
 
@@ -4366,6 +4384,12 @@
     int                  nvisinfo;
     XSetWindowAttributes attrib;
 
+    priv->ctx = Glib::MainContext::get_default ();
+    priv->mainloop = Glib::MainLoop::create (priv->ctx, false);
+    priv->sighupSource = CompSignalSource::create (SIGHUP, boost::bind (&PrivateScreen::handleSignal, priv, _1));
+    priv->sigintSource = CompSignalSource::create (SIGINT, boost::bind (&PrivateScreen::handleSignal, priv, _1));
+    priv->sigtermSource = CompSignalSource::create (SIGTERM, boost::bind (&PrivateScreen::handleSignal, priv, _1));
+
     dpy = priv->dpy = XOpenDisplay (name);
     if (!priv->dpy)
     {
@@ -4374,6 +4398,8 @@
 	return false;
     }
 
+    XSynchronize (dpy, TRUE);
+
 //    priv->connection = XGetXCBConnection (priv->dpy);
 
     snprintf (priv->displayString, 255, "DISPLAY=%s",

=== added file 'src/signalsource.cpp'
--- src/signalsource.cpp	1970-01-01 00:00:00 +0000
+++ src/signalsource.cpp	2012-01-21 13:34:31 +0000
@@ -0,0 +1,65 @@
+/*
+ * Copyright © 2012 Canonical, Ltd.
+ *
+ * Permission to use, copy, modify, distribute, and sell this software
+ * and its documentation for any purpose is hereby granted without
+ * fee, provided that the above copyright notice appear in all copies
+ * and that both that copyright notice and this permission notice
+ * appear in supporting documentation, and that the name of
+ * Canonical, Ltd. not be used in advertising or publicity pertaining to
+ * distribution of the software without specific, written prior permission.
+ * Canonical, Ltd. makes no representations about the suitability of this
+ * software for any purpose. It is provided "as is" without express or
+ * implied warranty.
+ *
+ * CANONICAL LTD. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
+ * NO EVENT SHALL Canonical, Ltd. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
+ * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
+ * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
+ * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ *
+ * Author: Sam Spilsbury <sam.spilsbury@canonical.com>
+ */
+
+#include "privatesignalsource.h"
+#include <glib-unix.h>
+
+CompSignalSource *
+CompSignalSource::create (int signum, const callbackFunc &f)
+{
+    return new CompSignalSource (signum, f);
+}
+
+CompSignalSource::CompSignalSource (int signum, const callbackFunc &f) :
+    mFunc (f),
+    mSignal (signum),
+    mSource (g_unix_signal_add_full (G_PRIORITY_HIGH,
+				     signum,
+				     CompSignalSource::callback,
+				     this,
+				     CompSignalSource::destroyed))
+{
+}
+
+gboolean
+CompSignalSource::callback (gpointer user_data)
+{
+    CompSignalSource *self = static_cast <CompSignalSource *> (user_data);
+    self->mFunc (self->mSignal);
+    return TRUE;
+}
+
+void
+CompSignalSource::destroyed (gpointer user_data)
+{
+    CompSignalSource *self = static_cast <CompSignalSource *> (user_data);
+    self->mSource = 0;
+}
+
+CompSignalSource::~CompSignalSource ()
+{
+    if (mSource)
+	g_source_remove (mSource);
+}

